name = "pyxem"
version = "0.21.0"
author = "Duncan Johnstone, Phillip Crout, Carter Francis, Magnus Nord"
copyright = "Copyright 2016-2025, The pyxem developers"
# Contributors listed by original committer, maintainers, then other
# within groups organised by  number of commits.
# See: https://github.com/pyxem/pyxem/graphs/contributors
credits = [
    "Duncan Johnstone",
    "Phillip Crout",
    "Carter Francis",
    "Magnus Nord",
    "Joonatan Laulainen",
    "Simon Høgås",
    "Eirik Opheim",
    "Eric Prestat",
    "Ben Martineau",
    "Tina Bergh",
    "Viljar Johan Femoen",
    "Niels Cautaerts",
    "Håkon Wiik Ånes",
    "Stef Smeets",
    "Andrew Ross",
    "Joseph Broussard",
    "Sivert Dagenborg",
    "Shuoyuan Huang",
    "Sean Collins",
    "Tom Furnival",
    "Daen Jannis",
    "Ida Hjorth",
    "Endre Jacobsen",
    "Mohsen Danaie",
    "Andrew Herzing",
    "Tim Poon",
    "Ruben Bjørge",
    "Tiarnan Doherty",
    "Affan Iqbal",
    "Jedrzej Morzy",
    "Tomas Ostasevicius",
    "Tor Inge Thorsen",
    "Matt von Lany",
    "Rob Tovey",
    "Petr Vacek",
]
license = "GPLv3+"
maintainer = "Duncan Johnstone, Phillip Crout, Magnus Nord, Carter Francis"
email = "pyxem.team@gmail.com"
status = "development"
